﻿#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# [SNIPPET_NAME: Script-Fu Composer]
# [SNIPPET_CATEGORIES: PyGTK, GimpFu]
# [SNIPPET_DESCRIPTION: Complete RAD tool for Gimp to build new script with currently exists]
# [SNIPPET_AUTHOR: Jeremy Czajkowski <jeremy.cz@wp.pl>]
# [SNIPPET_LICENSE: Creative commons <CC-BY-SA>]
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

import time
import os
import re

from gimpfu import *
import gimpenums

import gobject
import gtk.gdk
import pygtk

import json
import ConfigParser
import unicodedata
from random import randint

pygtk.require('2.0')


# Dictionaries are sometimes found in other languages as associative memories or associative arrays.
SC_I18N = {}
SC_PREF = {}
SC_VALS = {}
SC_FILES = {}
SC_FUNCMEMO = {}
SC_METHODS = {}


# Files & Directories
SC_ABSDIR = os.path.dirname(os.path.abspath(__file__)).lower()
SC_RESDIR = os.path.join(SC_ABSDIR , "_scfiles")
SC_LANGFILE = ""
SC_CONFIGFILE = ""


# Variables
SC_GIMPVERSION = ""
SC_TITLE = "Batch Script Composer"
SC_SAVEMODE_WARNING = "Work in safe mode. Choose you GIMP language correctly /not based on system settings, for example/."

# Array with one dimension
SC_EXCLUDED_TYPE = ['run-mode' , 'image' , 'layer' , 'drawable']
SC_BUTTONS = []
SC_PROCS = []


# Constants
SC_SECTION_HEADER = "USERDEF"
SC_PARAM_NOT_USED = "NOT_USED"
SC_ADD_OPENED_IMAGES = 1001
SC_ADD_RESOURCES = 1002
SC_CLEAR_LIST = 1003



class ResourceChooserButton(gtk.Fixed):
# @see http://stackoverflow.com/questions/6768779/test-filename-with-regular-expression
# @see http://stackoverflow.com/questions/169008/regex-for-parsing-directory-and-filename
    BG_PATH = os.path.join(SC_RESDIR, "images", "resource_chooser.png")
    KIND = ["dirname", "filename", "resname"]
    REGEX = {"filename": "^[\w\s-]+\.[A-Za-z]{3}$", "dirname": "^(.+)/([^/]+)$" , "resname":"\w"}
    ICON = {"filename": gtk.STOCK_FILE, "dirname": gtk.STOCK_DIRECTORY, "resname": gtk.STOCK_INDEX}

    ICON_SIZE = 16
    MENU_LIMIT = 20
    PADDING = 5

    def __init__(self, res_kind, win_title=None):
        gtk.Fixed.__init__(self)

        self.window_title = win_title
        self.orginal_kind = res_kind
        self.resource_kind = res_kind

        background = gtk.Image()
        background.set_from_file(self.BG_PATH)
        self.width = background.get_pixbuf().get_width()
        self.height = background.get_pixbuf().get_height()

        self.resource = gtk.Entry()
        self.resource.set_max_length(255)
        self.resource.connect("changed", self.changed)

        self.action = gtk.Image()
        try:
            self.action.set_from_stock(self.ICON[res_kind], 2)
        except:
            self.action.set_from_stock(self.ICON["resname"], 2)
            self.resource_kind = "resname"
            pass

        action_box = gtk.EventBox()
        action_box.add(self.action)
        action_box.add_events(gtk.gdk.BUTTON_RELEASE_MASK)
        action_box.connect("button-release-event", self.release)

        self.add(background)
        self.put(self.resource, self.PADDING, self.PADDING)
        self.put(action_box, self.width - self.ICON_SIZE - self.PADDING * 2, int(self.height * 0.5 - self.ICON_SIZE * 0.5))
        self.set_size_request(self.width, self.height)
        self.show_all()
        self.menu = self.create_menu()

    def create_menu(self):
      try:
        if "pattern" == self.orginal_kind:
          _, res_list = pdb.gimp_patterns_get_list("")
        elif "gradient" == self.orginal_kind:
          _, res_list = pdb.gimp_gradients_get_list("")
        else:
          res_list = []

        sort_list = []
        sort_list.extend(res_list)
        sort_list.sort()
        root = gtk.Menu()
        menu = gtk.Menu()
        mitem = gtk.MenuItem("#")
        root.append(mitem)
        mitem.show()
        mitem.set_submenu(menu)

        num = 0
        for res in sort_list:
          f_res = re.sub("[^\w# ]", "", res)
          child = gtk.MenuItem(f_res)
          if num < self.MENU_LIMIT:
            num = num + 1
          else:
            num = 0
            menu = gtk.Menu()
            mitem.set_label(mitem.get_label() + " - " + res[0])
            mitem = gtk.MenuItem(res[0])
            root.append(mitem)
            mitem.show()
            mitem.set_submenu(menu)

          menu.append(child)
          child.show()
          child.connect("activate", self.menuitem_activate)

        mitem = root.get_children()[0]
        submenu = mitem.get_submenu()
        if submenu is not None:
          first = submenu.get_children()[0]
          last = submenu.get_children()[submenu.get_children().__len__() - 1]
          if first is not None:
            f_letter = first.get_label()[0]
            l_letter = last.get_label()[0]
            mitem.set_label(f_letter + " - " + l_letter)
      except Exception, e:
          menu = gtk.Menu()
          mitem = gtk.MenuItem("Resource Chooser say: {0}".format(e))
          menu.append(mitem)
          mitem.show()
          return menu
      else:
          return root

    def get_directory(self):
        response = ""
        dialog = gtk.FileChooserDialog(self.window_title, None,
          gtk.FILE_CHOOSER_ACTION_OPEN,
          (
                  gtk.STOCK_CANCEL,
                  gtk.RESPONSE_CANCEL,
                  gtk.STOCK_ADD,
                  gtk.RESPONSE_OK
          )
        )
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)

        _filter = gtk.FileFilter()
        _filter.set_name(i18n_const("filter_local_folder"))
        _filter.add_pattern(".")
        dialog.add_filter(_filter)

        if (dialog.run() == gtk.RESPONSE_OK) :
                response = dialog.get_filename()
        dialog.destroy()

        return response

    def get_resource(self):
        return self.resource.get_text()

    def set_resource(self, res_kind, data):
        if res_kind in self.KIND:
            self.resource.set_text(data)

    def set_title(self, title):
        self.window_title = title.capitalize()

    def release(self, widget, event):
        if self.resource_kind == "dirname":
            outDirectory = re.sub("\\\\", "/", self.get_directory())
            self.resource.set_text(outDirectory)

        if self.resource_kind == "resname":
            self.menu.popup(None, None, None, 0, 0)

    def changed(self, entry):
        if re.match(self.REGEX[self.resource_kind], entry.get_text()):
            entry.set_icon_from_stock(0, None)
        else:
            entry.set_icon_from_stock(0, gtk.STOCK_DIALOG_ERROR)

    def menuitem_activate(self, widget):
        self.resource.set_text(widget.get_label())


class ProcedureChooserDialog(gtk.Dialog):

    def __init__(self):
        gtk.Dialog.__init__(self
                            , i18n_const("pdb_dialog")
                            , None
                            , gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                             gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))

        self.set_default_response(gtk.RESPONSE_CANCEL)

        # initiate the variable which will be returned after closing
        self._response = ""

        # create a TreeStore with one string column to use as the model
        self.treestore = gtk.TreeStore(str)

        # iterate over procedure database
        gpiter = self.treestore.append(None , [i18n_const("pdb_buildin")])
        opiter = self.treestore.append(None , [i18n_const("pdb_imported")])

        for name in SC_PROCS:
            if name.startswith("gimp-"):
                self.treestore.append(gpiter , [name])
            else:
                self.treestore.append(opiter , [name])

        # create the TreeView using treestore
        self.treeview = gtk.TreeView(self.treestore)

        # create the TreeViewColumn to display the data
        self.tvcolumn = gtk.TreeViewColumn()

        # add tvcolumn to treeview
        self.treeview.append_column(self.tvcolumn)

        # create a CellRendererText to render the data
        self.cell = gtk.CellRendererText()

        # add the cell to the tvcolumn and allow it to expand
        self.tvcolumn.pack_start(self.cell, True)

        # set the cell "text" attribute to column 0 - retrieve text from that column in treestore
        self.tvcolumn.add_attribute(self.cell, 'text', 0)

        # make it searchable
        self.treeview.set_search_column(0)

        # Allow sorting on the column
        self.tvcolumn.set_sort_column_id(0)
        self.tvcolumn.clicked()

        # Add vertical scrollbar
        swH = gtk.ScrolledWindow()
        swH.set_size_request(300 , 300)
        swH.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        swH.add(self.treeview)

        # Configure place for display informations about selected procedure(item)
        self.prochelp = gtk.Label(i18n_const("pdb_dialog_info"))
        self.prochelp.set_line_wrap(True)
        self.prochelp.set_size_request(280, 300)
        self.prochelp.set_alignment(xalign=0.2, yalign=0.1)

        # Add to main container
        hbox = gtk.HBox(True)
        hbox.add(swH)
        hbox.add(self.prochelp)

        # # Add scrolled window with tree view
        self.vbox.add(hbox)

        # # Actions
        self.treeview.connect('row-activated', self.row_activated)
        self.treeview.connect('move-cursor', self.row_activated)

        # # Show all widgets
        self.show_all()

    def row_activated(self, treeview, arg1, arg2):
        (model, seliter) = treeview.get_selection().get_selected()

        name_of_data = model.get_value(seliter, 0)
        try:
            res = pdb.query(name_of_data)
            if res:
                proc = pdb[res[0]]
        # choose between full description and tooltip
                proc_description = proc.proc_help if proc.proc_help.__len__() > 1 else proc.proc_blurb.capitalize()
        # counting words to avoid funny text
        # http://stackoverflow.com/questions/17507876/trying-to-count-words-in-a-string-python
                words = re.split(r'[^0-9A-Za-z]+', proc_description)
                if words.__len__() <= 4:
                    proc_description = proc.proc_blurb.capitalize()
                formated_information = "%s\r\n\n%s" % (proc.proc_author, proc_description)

                self.prochelp.set_text(formated_information)
                self._response = name_of_data
        except Exception , ex:
            pdb.gimp_message("Procedure Chooser say: {0}".format(ex))
            pass

    def get_selected(self):
        return self._response


class BatchScriptComposer(gtk.Window):

    WIDGET_BORDER_SIZE = 10
    FRAME_BORDER_COLOR = "blue"

    WIDTH = 650
    HEIGHT = 450

    def __init__(self, safemode):
        gtk.Window.__init__(self)

        self.set_title(SC_TITLE)
        self.set_resizable(False)
        self.set_border_width(self.WIDGET_BORDER_SIZE / 2)
        self.set_size_request(self.WIDTH , self.HEIGHT)

        self.button_x = 1
        self.name_of_previous = ""
        self.output_directory = ""

        self.set_events(gtk.gdk.ALL_EVENTS_MASK)
        self.connect("delete-event" , self.delete_event)
        self.connect("destroy" , self.destroy_event)

        # The INI file format parser
        self.config = ConfigParser.ConfigParser()

        # ToolTip Instance
        self.tooltips = gtk.Tooltips()

        # Main layout
        frame_set = gtk.Frame(i18n_const("frame_1"))
        frame_set.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        frame_set.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse(self.FRAME_BORDER_COLOR))
        frame_set.set_size_request(self.WIDTH , 96)

        frame_options = gtk.Frame(i18n_const("frame_2"))
        frame_options.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        frame_options.modify_bg(gtk.STATE_NORMAL , gtk.gdk.color_parse(self.FRAME_BORDER_COLOR))

        self.messagebox = gtk.Label(SC_SAVEMODE_WARNING)

        hbox = gtk.HBox(True)
        vbox = gtk.VBox()
        vbox.pack_start(self.messagebox , False , False)
        vbox.pack_start(frame_set , False)
        vbox.add(hbox)

        # Files
        toolbar = gtk.Toolbar()
        toolbar.set_orientation(gtk.ORIENTATION_HORIZONTAL)
        toolbar.set_style(gtk.TOOLBAR_BOTH_HORIZ)

        stock_add = gtk.image_new_from_stock(gtk.STOCK_ADD , gtk.ICON_SIZE_SMALL_TOOLBAR)
        stock_clear = gtk.image_new_from_stock(gtk.STOCK_CLEAR , gtk.ICON_SIZE_SMALL_TOOLBAR)
        stock_convert = gtk.image_new_from_stock(gtk.STOCK_CONVERT , gtk.ICON_SIZE_SMALL_TOOLBAR)

        toolbar.append_item(i18n_const("button_1") , None , "Private" , stock_add , self.toolbar_action , SC_ADD_RESOURCES)
        toolbar.append_item(i18n_const("button_2") , None , "Private" , stock_convert , self.toolbar_action , SC_ADD_OPENED_IMAGES)
        toolbar.append_item(i18n_const("button_3") , None , "Private" , stock_clear , self.toolbar_action , SC_CLEAR_LIST)
        toolbar.show_all()

        mn = gtk.TreeViewColumn(None , gtk.CellRendererText() , text=0)

        self.tree = gtk.TreeView(gtk.ListStore(gobject.TYPE_STRING))
        self.tree.set_headers_visible(False)
        self.tree.append_column(mn)

        swH = gtk.ScrolledWindow()
        swH.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        swH.add(self.tree)

        box_files = gtk.VBox()
        box_files.pack_start(toolbar , False , False , 5)
        box_files.pack_end(swH , True , True)

        hbox.add(box_files)
        hbox.add(frame_options)

        # Options
        box_options = gtk.VBox()

        self.button_output = gtk.MenuToolButton(None , i18n_const("option_output_button"))
        self.button_output.connect("clicked", self.output_action)

        self.box_params = gtk.VBox()

        swH = gtk.ScrolledWindow()
        swH.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        swH.add_with_viewport(self.box_params)

        swH.get_child().set_shadow_type(gtk.SHADOW_NONE)

        box_options.set_border_width(self.WIDGET_BORDER_SIZE)
        box_options.pack_start(self.button_output , False , False)
        box_options.pack_end(swH)
        box_options.show_all()

        frame_options.add(box_options)

        # Sets
        self.button_addset = gtk.Button(None , gtk.STOCK_ADD)
        self.button_addset.set_size_request(72 , 72)
        self.button_addset.set_image_position(gtk.POS_TOP)
        self.button_addset.connect("clicked", self.addset_action)

        self.container = gtk.Fixed()
        self.container.add(self.button_addset)

        frame_set.add(self.container)

        # Action Area of Manipulation Set
        button_box = gtk.HButtonBox()
        button_box.set_layout(gtk.BUTTONBOX_END)
        button_box.set_spacing(10)

        button = gtk.Button(stock=gtk.STOCK_OPEN)
        _button = gtk.Button(stock=gtk.STOCK_REMOVE)
        self.button_save = gtk.Button(stock=gtk.STOCK_SAVE_AS)
        self.button_execute = gtk.Button(stock=gtk.STOCK_EXECUTE)

        self.button_save.set_state(gtk.STATE_INSENSITIVE)
        self.button_execute.set_state(gtk.STATE_INSENSITIVE)

        button_box.add(button)
        button_box.add(self.button_save)
        button_box.add(_button)
        button_box.add(self.button_execute)

        button.connect("clicked", self.saveload, gtk.FILE_CHOOSER_ACTION_OPEN)
        _button.connect("clicked", self.clearset_action)
        self.button_save.connect("clicked", self.saveload, gtk.FILE_CHOOSER_ACTION_SAVE)
        self.button_execute.connect("clicked", self.execute_action)

        vbox.pack_end(button_box , False)

        # Progress
        self.pbar = gtk.ProgressBar()
        vbox.pack_end(self.pbar , False)

        # Show content of main layout
        self.add(vbox)
        self.show_all()
        if not safemode:
            self.messagebox.hide()

    def main(self):
        """
        This is a recent change in GTK - the developers wanted icons not to appear on buttons.
          On Linux, this can be changed by editing the gconf key
          On windows, set a value in your gtkrc file and use a theme that supports icons

        The Python equivalent for setting the property without having to change any system configuration files is:
        """
        settings = gtk.settings_get_default()
        settings.props.gtk_button_images = True

        """
        Start GTK main loop
        """
        gtk.main()

    def delete_event(self, widget, event, data=None):
        return False

    def destroy_event(self, widget, data=None):
        gtk.main_quit()

    def addset_action(self, widget):
        dialog = ProcedureChooserDialog()
        if (dialog.run() == gtk.RESPONSE_ACCEPT) :
            response = dialog.get_selected()
            if response:
                try:
                    button = self.create_buttonandmenu(response)
                except Exception, ex:
                    pdb.gimp_message("{0}[addset_action] say: {1}".format(self.__class__.__name__ , ex))
                    pass
                else:
                    self.button_save.set_sensitive(True)
                    self.button_execute.set_sensitive(True)

                    event = gtk.gdk.Event(gtk.gdk.KEY_PRESS)
                    event.keyval = 32
                    self.button_press(button, event)
                    button.grab_focus()

        dialog.destroy()

    def clearset_action(self, widget):
        try:
            self.button_addset.hide()
            for child in self.container.get_children():
                if child.get_visible():
                    self.container.remove(child)
                    SC_BUTTONS.remove(child)
                    del child

            for child in self.box_params.get_children():
                self.box_params.remove(child)
                del child

            self.button_x = 1
            self.container.move(self.button_addset , 1 , 0)
            self.button_addset.show()
        except Exception , ex:
            pdb.gimp_message("{0}[clearset_action] say: {1}".format(self.__class__.__name__ , ex))
            pass
        else:
            self.button_execute.set_sensitive(False)
            self.button_save.set_sensitive(False)

    def menuitem_remove_action(self , menuitem , button):
        try:
            index = SC_BUTTONS.index(button)
        except:
            pass
        else:
            for num in range(index, SC_BUTTONS.__len__()):
                x = self.container.child_get_property(SC_BUTTONS[num] , "x")
                self.container.move(SC_BUTTONS[num] , x - 72 , 0)

            SC_VALS[button.name_of_data] = []
            for child in self.box_params.get_children():
                self.box_params.remove(child)
                del child

            self.container.remove(button)
            SC_BUTTONS.remove(button)
            del button

            self.container.move(self.button_addset , x , 0)
            self.button_x = x

    def button_press(self, widget, event):
        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
            widget.popup(None, None, None, event.button, event.time)
            pass
        if (event.type == gtk.gdk.KEY_PRESS and event.keyval in [32, 65293]) or (event.type == gtk.gdk.BUTTON_PRESS and event.button == 1):
            try:
                message_label = gtk.Label(i18n_const("pdb_without_input_params"))
                self.collect_data(True)

                # Name of class of event trigger
                cname = widget.__class__.__name__.lower()
                # Choose resource
                if cname == 'menu':
                    menuitem = widget.get_children()[0]
                    name_of_data = menuitem.name_of_data
                elif cname == 'button':
                    name_of_data = widget.name_of_data
                else:
                    return
                self.name_of_previous = name_of_data

                # Choose the array of values to put for editing
                values = SC_VALS[name_of_data]
                nvalues = 0

                proc_name = name_of_data[:name_of_data.rfind('-')]                
                if pdb[proc_name].nparams > 0:
                # Parameter's Collection
                    params = pdb[proc_name].params
                # Collect number of editable parameters
                    nvparams = pdb[proc_name].nparams
                    for p in params:
                        size_request = False
                        p_kind = p[0]
                        p_type = p[1].lower()
                        p_length = 0 if p_type in ['toggle'] else 16
                        p_label = text2SuitableTip(p[2], p_length) if p_type in ['value', 'toggle', 'enum', 'option', 'string'] else value2SuitableLabel(p[1])
                    # #RUN-MODE
                        if p_type == 'run-mode':
                            nvparams -= 1
                            continue
                    # #IMAGE
                        if p_type == 'image':
                            nvparams -= 1
                            entry = gtk.HSeparator()
                            size_request = True
                    # #LAYER, DRAWABLE
                        elif p_type in ['layer', 'drawable']:
                            nvparams -= 1
                            entry = gtk.HSeparator()
                            size_request = True
                    # #TOGGLE
                        elif p_type == 'toggle':
                            entry = gtk.CheckButton(p_label)
                            if values:
                                entry.set_active(values[nvalues])
                    # #ENUM
                        elif p_type == 'enum':
                            combostore = gtk.ListStore(gobject.TYPE_INT , gobject.TYPE_STRING)
                            name = p[2].upper()
                            for key in gimpenums.__dict__:
                                if key.startswith(name):
                                    label = key[name.__len__() + 1:]
                                    combostore.append([gimpenums.__dict__[key] , label])
  
                            entry = gtk.ComboBox(combostore)
                            cell = gtk.CellRendererText()
                            entry.pack_start(cell, True)
                            entry.add_attribute(cell, 'text', 1)
                            size_request = True
                            if values:
                                entry.set_active(values[nvalues])
                    # #OPTION
                        elif p_type == 'option':
                            entry = gtk.ComboBox()
                            size_request = True
                            if values:
                                entry.set_active(values[nvalues])
                    # #COLOR
                        elif p_type == 'color' or p_kind == 10:
                            entry = gtk.ColorButton()
                            size_request = True
                            entry.set_title(p_label)   
                            if values:
                                val = values[nvalues]
                                try:
                                    mycolor = gtk.gdk.color_parse(val) if val.__class__.__name__.lower() == 'str' else val
                                    entry.set_color(mycolor)
                                except:
                                    entry.set_color(gtk.gdk.Color())
                    # #FONT
                        elif p_type == 'font':
                            entry = gtk.FontButton()
                            entry.set_title(p_label)
                            entry.set_show_style(False)
                            if values:
                                entry.set_font_name(values[nvalues])
                    # #GRADIENT
                        elif p_type == 'gradient':
                            entry = ResourceChooserButton(p_type)
                            if values:
                                entry.set_resource("resname", values[nvalues])
                    # #PATTERN
                        elif p_type == 'pattern':
                            entry = ResourceChooserButton(p_type)
                            if values:
                                entry.set_resource("resname", values[nvalues])
                    # #DIRECTORY
                        elif p_type == 'dirname':
                            entry = ResourceChooserButton(p_type)
                            if values:
                                entry.set_resource("dirname", values[nvalues])
                    # #FILENAME
                        elif p_type == 'filename' or p_type.endswith("-filename"):
                            entry = ResourceChooserButton('filename')
                            if values:
                                entry.set_resource("filename", values[nvalues])
                    # #OTHERS
                        else:
                            entry = gtk.Entry()
                            size_request = True
                            if values and values[nvalues] is not None:
                                entry.set_text("{0}".format(values[nvalues]))
                # Add additional widgets
                        hbox = gtk.HBox()
                        hbox.pack_start(entry , False)
                        if not p_type in ['toggle']:
                            hbox.pack_end(gtk.Label(p_label))
                # Add tooltip
                        if p[2]:
                            self.tooltips.set_tip(entry, text2SuitableTip(p[2]))
                # Set properties
                        entry.hide_on_delete()
                        if size_request:
                            entry.set_size_request(150, 22)
                        if not p_type in SC_EXCLUDED_TYPE:
                            hbox.show_all()
                # Pack widgets for chosen parameter
                        self.box_params.pack_start(hbox , False , False , 5)
                # Increment counter
                        if nvalues < values.__len__():
                            nvalues += 1
                # Display message
                    if nvparams < 1:
                        self.box_params.add(message_label)
                        message_label.show()
                else:
                # Display message
                    self.box_params.add(message_label)
                    message_label.show()
            except Exception, ex:
                pdb.gimp_message("{0}[button_press] say: {1}".format(self.__class__.__name__ , ex))

    def output_action(self, widget):
        patch = get_paths(i18n_const("option_output_dialog") , gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER , False)
        if patch:
            self.output_directory = patch[0]
            self.button_output.set_label(os.path.split(patch[0])[1])

    def toolbar_action(self, widget, data):
        model = self.tree.get_model()
        if data == SC_ADD_RESOURCES:
            resList = get_paths(i18n_const("button_1") , gtk.FILE_CHOOSER_ACTION_OPEN , True)
            if resList:
                for filepath in resList:
                    if os.path.isdir(filepath):
                        for name in os.listdir(filepath):
                            srcpath = os.path.join(filepath , name)
                            if os.path.isfile(srcpath) and not SC_FILES.has_key(name):
                                SC_FILES[name] = srcpath
                                model.append([name])
                    else:
                        filename = os.path.basename(filepath)
                        if not SC_FILES.has_key(filename):
                            SC_FILES[filename] = filepath
                            model.append([filename])
        elif data == SC_ADD_OPENED_IMAGES:
            for image in gimp.image_list():
                filepath = pdb.gimp_image_get_filename(image)
                filename = os.path.basename(filepath)
                if not SC_FILES.has_key(filename):
                    SC_FILES[filename] = filepath
                    model.append([filename])
        elif data == SC_CLEAR_LIST:
            model.clear()
            SC_FILES.clear()

    def execute_macro(self,macro,value,filepath):
        val = value if value else filepath
        
        for prop in macro.split(';'):
            dbloc = prop.find('@')
            dbpath= None
            if dbloc > 1:
                dbpath = prop[dbloc+1:].strip()
                prop = prop[:dbloc].strip()

            prop_args = prop.split(',') 
            if dbpath:
                SC_FUNCMEMO[prop_args[0]] = dbpath
                       
            if SC_METHODS.has_key(prop_args[0]):
                prop_result = SC_METHODS[prop_args[0]](val)
                if prop_args.__len__() > 1:
                    val = prop_result[int(prop_args[1])]
                else:
                    val = prop_result
        
        return val

    def execute_action(self, widget):

        def is_empty(data):
            result = data is None
            if not result:
                if data.__class__.__name__ == 'str':
                    result = data.strip().__len__() == 0
            return result

        def is_accessable(data):
            result = False
            if data.__class__.__name__ == 'str':
                result = data.strip() == SC_PARAM_NOT_USED
            return result

        nfiles = 0
        self.pbar.set_fraction(0.00)
        if os.path.isdir(self.output_directory):
            try:
                self.collect_data(False)
                for filename , filepath in SC_FILES.iteritems():
                    progress = round(abs(float(nfiles) / SC_FILES.__len__()) , 2)

                    self.pbar.set_fraction(progress)
                    self.pbar.set_text(filename)
                    nfiles -= 1
                    vals = {}
                    try:
                        image = pdb.gimp_file_load(filepath , filename)
                    except Exception , ex:
                        pdb.gimp_message("Loader say: {0}".format(ex))
                        pass
                    else:
                        layer = pdb.gimp_image_get_active_layer(image)
                        destpath = os.path.join(self.output_directory , filename)

                        for button in SC_BUTTONS:
                            name = button.name_of_data[:button.name_of_data.rfind('-')]
                            item = SC_VALS[button.name_of_data]

                            args = []
                            num = 0
                            for param in pdb[name].params:
                            # #PROLOG
                                class_name = item[num].__class__.__name__.lower()
                                if class_name == 'color':
                                    item[num] = item[num].to_string()
                            # #IMAGE
                                if param[0] == 13:
                                    args.append(image)
                            # #LAYER
                                elif param[0] == 14:
                                    if is_accessable(item[num]):
                                        args.append(layer)
                                    else:
                                        args.append(item[num])
                            # #DRAWABLE
                                elif param[0] == 16:
                                    args.append(pdb.gimp_image_get_active_drawable(image))
                            # #PREVIOUS
                                elif vals.has_key(param[0]):
                                    if param[0] in [0, 3, 4]:
                                        text = item[num]
                                        if is_empty(text):
                                            val = vals[param[0]]
                                        elif text.find(';') > 1:
                                            val = self.execute_macro(text, vals[param[0]], filepath)
                                        else:
                                            val = text
                                    else:
                                        val = vals[param[0]]
                                    args.append(val)
                                else:
                            # #CURRENT
                                    val = item[num]
                                    if val.__class__.__name__ == 'str' and val.find(';') > 1:
                                        val = self.execute_macro(val,'',filepath)

                                    args.append(val)

                                if num < item.__len__():
                                    num += 1

                            rv = pdb[name].return_vals
                            if rv:
                                result = pdb[name](*args)
                                vals.clear()
                                if rv.__len__() == 1:
                                    if rv[0][0] == 14:
                                        layer = result
                                    vals[rv[0][0]] = result
                            else:
                                pdb[name](*args)

                        try:
                            pdb.gimp_file_save(image , layer , destpath , filename)
                        except Exception , ex:
                            pdb.gimp_message("Saver say: \n\t{1}\n\t{0}".format(ex, ex.__class__.__name__))
                            pass
                        pdb.gimp_image_delete(image)
                        
                    while gtk.events_pending():
                        gtk.main_iteration()

            except Exception , ex:
                pdb.gimp_message("Executor say: \n\t{1}\n\t{0}".format(ex, ex.__class__.__name__))
                pass
            else:
                self.pbar.set_fraction(1.00)

    def create_buttonandmenu(self , proc_name):
        button = gtk.Button()
        button.name_of_data = "{0}-{1}".format(proc_name , randint(1 , int(time.time() / 1000)))

        button.set_image(gtk.image_new_from_stock(gtk.STOCK_EXECUTE , gtk.ICON_SIZE_BUTTON))
        button.set_size_request(72 , 72)
        button.hide_on_delete()
        button.show()

        menu = gtk.Menu()

        item_2 = gtk.MenuItem(i18n_const("menuitem_removeme"))
        item_1 = gtk.MenuItem(proc_name)
        item_1.name_of_data = button.name_of_data

        menu.append(item_1)
        menu.append(item_2)

        item_1.set_state(gtk.STATE_INSENSITIVE)
        item_2.connect("activate", self.menuitem_remove_action , button)

        menu.show_all()

        SC_BUTTONS.append(button)
        SC_VALS[button.name_of_data] = []

        button.connect_object("event" , self.button_press , menu)

        self.button_x += 72
        self.container.move(self.button_addset , self.button_x , 0)
        self.container.put(button , self.button_x - 72 , 0)
        button.grab_focus()

        self.tooltips.set_tip(button, pdb[proc_name].proc_blurb)
        return button

    def saveload(self, widget, action):
        paths = get_paths(i18n_const("saveload_dialog") , action , False)
        if paths:
            filepath = os.path.splitext(paths[0])[0] + ".bimp".lower()
            try:
                if action == gtk.FILE_CHOOSER_ACTION_SAVE:
                    block_counter = 0
                    filename = slugify(os.path.basename(filepath))
                    filepath = os.path.join(os.path.dirname(filepath) , filename)
                    with open(filepath, "w") as f:
                        f.write("#BIMP compatible\n")
                        f.write("#MANIPULATION SET DEFINITION\n")
                        for btn in SC_BUTTONS:
                            block_name = btn.name_of_data
                            proc_name = block_name[:block_name.rfind('-')]
                            f.write("\n[%s%d]\n" % (SC_SECTION_HEADER,block_counter))
                            f.write("procedure=%s\n" % proc_name)
                            f.write("num_params=%d\n" % pdb[proc_name].nparams)
                            block_counter += 1
                            param_counter = 0
                            val_counter = 0
                            params = pdb[proc_name].params
                            for p in params:
                                if val_counter <= SC_VALS[block_name].__len__():
                                    val = SC_VALS[block_name][val_counter]
                                    val_counter += 1

                                    if p[1].lower() == 'color' or p[0] == 10:
                                        if hasattr(val, 'to_string'):
                                            val = val.to_string().upper()
                                f.write("PARAM{0}={1}\n".format (param_counter , val))
                                param_counter += 1
                    f.close()

                if action == gtk.FILE_CHOOSER_ACTION_OPEN:
                    self.config.read(filepath)
                    for section in self.config.sections():
                     if section.startswith(SC_SECTION_HEADER):
                       # Create button
                           proc_name = self.config.get(section , "procedure")
                           name_of_data = self.create_buttonandmenu(proc_name).name_of_data

                       # Read property values
                           SC_VALS[name_of_data] = []
                           nparams = self.config.get(section , "num_params")
                           if nparams:
                               for num in range(int(nparams)):
                                   try:
                                       val = self.config.get(section , "PARAM%d" % int(num))
                                   except (ValueError , TypeError) , ex:
                                       pdb.gimp_message("Loader say: {0}".format(ex))
                                       val = ""
                                       pass
                                   else:
                                       if val == 'None':
                                           val = None
                                       else:
                                           val = int(val) if val.isdigit() else val

                                   SC_VALS[name_of_data].append(val)


            except Exception , ex:
                pdb.gimp_message("{0}[saveload] say: {1}".format(self.__class__.__name__ , ex))
                pass

            if SC_BUTTONS.__len__() > 0:
                event = gtk.gdk.Event(gtk.gdk.KEY_PRESS)
                event.keyval = 32
                self.button_press(SC_BUTTONS[0], event)
                SC_BUTTONS[0].grab_focus()

            self.button_execute.set_sensitive(True)
            self.button_save.set_sensitive(True)

    def collect_data(self, cleanup):
        SC_VALS[self.name_of_previous] = []
        for child in self.box_params.children():
        # Data instance
            entry = child.get_children()[0] if hasattr(child , "get_children") else child
            class_name = entry.__class__.__name__.lower()
        # Collect data
            if class_name == "entry":
                text = entry.get_text().strip()
                SC_VALS[self.name_of_previous].append(text if text else None)
            elif class_name == "combobox":
                index = 0 if entry.get_active() < 0 else entry.get_active()
                SC_VALS[self.name_of_previous].append(index)
            elif class_name == "resourcechooserbutton":
                SC_VALS[self.name_of_previous].append(entry.get_resource())
            elif class_name == "checkbutton":
                SC_VALS[self.name_of_previous].append((0, 1)[entry.get_active()])
            elif class_name == "fontbutton":
                SC_VALS[self.name_of_previous].append(entry.get_font_name())
            elif class_name == "colorbutton":
                SC_VALS[self.name_of_previous].append(entry.get_color())
            else:
                SC_VALS[self.name_of_previous].append(SC_PARAM_NOT_USED)
        # Clear working area of Property Manager
            if cleanup:
                self.box_params.remove(child)
                del child

def i18n_const(key):
    if SC_I18N.has_key(key):
        return SC_I18N[key]
    else:
        return key

def value2SuitableLabel(data):
    return re.sub('[-_]', ' ', data).strip().lower().capitalize()

def text2SuitableTip(data , length=0):
    """
    Replace with pattern, remove white chars and capitalize
    """
    patterns = {"_":""}
    formated = data
    for arg1 , arg2 in patterns.iteritems():
        formated = re.sub(arg1 , arg2 , formated , 0 , re.IGNORECASE)

    formated = formated.strip().capitalize()
    ff = formated.split()
    if ff[0].__len__() < 4:
        ff[0] = ff[0].upper()
        formated = " ".join(ff)

    # If it is given the maximum length of the string that divide so that all the words were,
    # on the occasion of removing everything that is not a letter
    if length:
        nlabel = 0
        words = formated.split(" ")
        accepted = []
        for word in words:
            nlabel += word.__len__()
            new_word = re.sub('[^-a-zA-Z]', '', word).strip()
            if new_word:
                accepted.append(new_word)
            if nlabel > length:
                break

        formated = " ".join(accepted)

    return formated

def slugify(value):
    """
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.

    Based on code in: Django
    """
    value = unicodedata.normalize('NFKD', unicode(value))
    value = value.encode('ascii', 'ignore')
    value = re.sub('[^\w\s.-]', '', value).strip().lower()
    value = re.sub('[-\s]+', '-', value)
    return value

def multiple_replace(dict, text):
  """
  This recipe shows how to use the Python standard re module to perform
  single-pass multiple string substitution using a dictionary.

  http://code.activestate.com/recipes/81330-single-pass-multiple-replace
  """
  # Create a regular expression  from the dictionary keys
  regex = re.compile("(%s)" % "|".join(map(re.escape, dict.keys())))

  # For each match, look-up corresponding value in dictionary
  return regex.sub(lambda mo: dict[mo.string[mo.start():mo.end()]], text)


def get_paths(title, action , multiple):
    response = ""
    try:
        dialog = gtk.FileChooserDialog(
                title
                , None
                , action
                , (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                  gtk.STOCK_APPLY, gtk.RESPONSE_ACCEPT)
        )
        dialog.set_select_multiple(multiple)
        dialog.set_default_response(gtk.RESPONSE_ACCEPT)

        if action == gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER:
            _filter = gtk.FileFilter()
            _filter.set_name(i18n_const("filter_local_folder"))
            _filter.add_pattern(".")
            dialog.add_filter(_filter)
        elif not multiple:
            _filter = gtk.FileFilter()
            _filter.set_name(i18n_const("filter_bimp"))
            _filter.add_pattern("*.bimp")
            dialog.add_filter(_filter)

        if (dialog.run() == gtk.RESPONSE_ACCEPT) :
                response = dialog.get_filenames()
        dialog.destroy()
    except Exception , ex:
        pdb.gimp_message("{0}[get_paths] say: {1}".format(ex.__class__.__name__ , ex))
        pass
    else:
        return response

def build_local_db(storage_path):
  try:
    sf_const = "SF-OPTION"
    sf_prefix= "script-fu"
    temp_path = SC_PREF["temp-path"]

    for item in os.listdir(storage_path):
      db = {}
      opt= {}
      basename = os.path.splitext(item)
      dbfile = os.path.join(storage_path, item)
      if basename[1] == ".scm":
        registerLine = ""
        scriptName = ""
        constantFounded = False
        with open(dbfile, "r") as src:
          for line in src.readlines():
            line = line.lstrip()
            if line.startswith("(script-fu-register"):
              registerLine = line
              s_line = re.sub('[\s]+', ' ', line.strip()).split(" ")

              if s_line.__len__() > 1:
                scriptName = slugify(s_line[1])

            if registerLine and not scriptName and registerLine != line:
              re_result = re.match('^[\w"\'-]+$', line.strip())
              if re_result is not None:
                s_line = slugify(re_result.group(0))
                if s_line.startswith(sf_prefix):
                  scriptName = s_line

            if registerLine and scriptName:
                if not db.has_key(scriptName):
                    db[scriptName] = []
                if re.search(sf_const, line):
                    constantFounded = True

            if line.startswith("SF-") and not line.startswith(sf_const):
                constantFounded = False

            if constantFounded:
                db[scriptName].append(line)

          src.close()

        regexp = re.compile(r'"[a-zA-Z ]+"' ,re.IGNORECASE)
        for name in db:
            if db[name]:
                output = "\n".join(db[name])
                output = re.sub("[\s ]+"," ",output).replace('_','')
                output = output[output.find(sf_const):].strip()

                opt[name] = []
#                 for item in output.split(sf_const):
#                     pdb.gimp_message(item)

#         if opt:
#           for name in opt:
#             try:
#               locale_basename = "%s.json" % name.replace(sf_prefix ,'')
#               locale_file = os.path.join(temp_path, locale_basename)
#
#               if opt[name]:
#                   with open(locale_file, "w") as dest:
#                     json.dump(opt[name], dest)
#             except Exception, e:
#               pdb.gimp_message("build_local_db[%s] <= %s" % (locale_basename, e))
#               pass
  except Exception, ex:
    if not storage_path.strip():
      storage_path = 'empty'
    pdb.gimp_message( "build_local_db[%s]\n storage:\t %s\n temp:\t %s \n\n %s" % (ex.__class__.__name__,storage_path,temp_path,ex) )
    pass


def loadtext(srcpath):
    if SC_FUNCMEMO.has_key('loadtext'):
        if os.path.isfile(SC_FUNCMEMO['loadtext']):
            filepath = SC_FUNCMEMO['loadtext']
            srcname = os.path.splitext( os.path.basename(srcpath) )[0]
            with open(filepath,'r') as inifile:
                for line in inifile:
                    inidata = line.partition('=')
                    if inidata[0].strip() == srcname:
                        return inidata[2].strip()

    return 'Not described yet'


def runner():

    # Gimp preferences
    SC_GIMPVERSION = os.path.splitext(pdb.gimp_version())[0]
    SC_CONFIGFILE = os.path.join(os.environ["userprofile"] , ".gimp-" + SC_GIMPVERSION , "gimprc")

    if os.path.isfile(SC_CONFIGFILE) and SC_PREF.__len__() == 0:
        namepattern = "[a-zA-Z0-9_-]+[ ]+"
        regexp = re.compile(r'^\(%s.+\)$' % namepattern , re.IGNORECASE)
        with open(SC_CONFIGFILE , "r") as f:
            for line in f.readlines():
                line = line.strip()
                if regexp.match(line):
                    temp = re.search(namepattern, line[1:-1])
                    value = line[temp.end():-1].strip().replace('"', '').split(os.pathsep)

                    if value.__len__() == 1:
                        value = value[0]
                    SC_PREF[temp.group().strip()] = value

    # Methods
    SC_METHODS['basename'] = os.path.basename
    SC_METHODS['dirname']  = os.path.dirname
    SC_METHODS['splitext'] = os.path.splitext
    SC_METHODS['loadtext'] = loadtext

    # Collect Gimp procedures
    regex = re.compile(r'Use(.){1,}[\'"-a-z]+[\s]+instead.$', re.IGNORECASE)
    for name in gimp.pdb.query():
        if regex.search(gimp.pdb[name].proc_blurb):
            continue
        SC_PROCS.append(name)

    # Internationalization
    safemode = False
    if SC_PREF.has_key("language"):
        SC_PREF["language"] = SC_PREF["language"][:2]
    else:
        SC_PREF["language"] = "en"
        safemode = True
    SC_LANGFILE = os.path.join(SC_RESDIR , "locale" , SC_PREF["language"].lower() + ".json")
    SC_I18N.clear()
    if os.path.isfile(SC_LANGFILE):
        try:
            text = open(SC_LANGFILE , "r").read()
            text = text.decode("windows-1250")
            matrix = json.loads(text)
        except IOError:
            pass
        else:
            SC_I18N.update(matrix)
    for path in SC_PREF["script-fu-path"]:
        build_local_db(path)

    # Initialize plug-in
    BatchScriptComposer(safemode).main()
    return None


register(
         "python-fu-batch-scriptcomposer"
        , "Complete programming environment to automated Gimp operations"
        , "Complete programming environment to automated Gimp operations." +
          " Applies GIMP manipulations on groups of images. Deprecated procedures in your version of GIMP is excluded from use."
        , "Jeremy Czajkowski"
        , "Jeremy Czajkowski"
        , "September 2013"
        , "<Toolbox>/_Tools/" + SC_TITLE +"..."
        , None
        , []
        , []
        , runner)

if __name__ == "__main__":
    main()
